import typing as t

import numpy as np
import pandas as pd
from pandas.core.generic import NDFrame

ArrType = t.TypeVar("ArrType", np.ndarray, NDFrame)
ArrCallable = t.Callable[[ArrType], ArrType]
ColRef = t.TypeVar("ColRef", str, pd.Series)


def tabulate_one_way(
        x: pd.Series,
        values: t.Optional[pd.Series] = None,
        aggfunc: t.Optional[t.Union[ArrCallable, t.Iterable[ArrCallable]]] = None,
        **kwargs
) -> pd.DataFrame:
    if values is None and aggfunc is not None:
        raise ValueError("aggfunc cannot be used without values.")

    elif aggfunc is None and values is not None:
        raise ValueError("values cannot be used without an aggfunc.")

    elif values is None and aggfunc is None:
        values = pd.Series(1, index=x.index, name="_values")

        def percent(v):
            return np.round(np.size(v) / len(x) * 100, 2)

        aggfunc = [np.size, percent]

    if not isinstance(aggfunc, t.Iterable):
        aggfunc = [aggfunc]

    df = pd.concat([x, values], axis=1)
    return df.groupby(x.name, **kwargs)[values.name].agg(aggfunc)


def tabulate(
        x: t.Union[pd.Series, t.Iterable[pd.Series]],
        y: t.Optional[t.Union[pd.Series, t.Iterable[pd.Series]]] = None,
        values: t.Optional[pd.Series] = None,
        aggfunc: t.Optional[t.Union[ArrCallable, t.Iterable[ArrCallable]]] = None,
        **kwargs
) -> pd.DataFrame:
    if y is None:
        return tabulate_one_way(x, values=values, aggfunc=aggfunc, **kwargs)
    else:
        return pd.crosstab(x, y, values=values, aggfunc=aggfunc, **kwargs)


class BaseTabAccessor(object):

    def __init__(self, pandas_obj: pd.DataFrame):
        self._obj = pandas_obj

    def _lookup(
            self,
            c: t.Optional[t.Union[ColRef, t.Iterable[ColRef]]]
    ) -> t.Optional[t.Union[pd.Series, t.List[pd.Series]]]:
        if c is None or isinstance(c, pd.Series):
            return c
        else:
            if isinstance(c, str):
                return self._obj[c]
            elif isinstance(c, t.Iterable):
                return list(map(self._lookup, c))
            else:
                raise self._obj[c]


@pd.api.extensions.register_dataframe_accessor("tab")
class TabDataFrameAccessor(BaseTabAccessor):

    def __call__(
            self,
            x: t.Union[ColRef],
            y: t.Optional[t.Union[ColRef, t.Iterable[ColRef]]] = None,
            values: t.Optional[pd.Series] = None,
            aggfunc: t.Optional[t.Union[ArrCallable, t.Iterable[ArrCallable]]] = None,
            **kwargs
    ) -> pd.DataFrame:
        x = self._lookup(x)
        y = self._lookup(y)
        values = self._lookup(values)
        return tabulate(x, y, values=values, aggfunc=aggfunc, **kwargs)


@pd.api.extensions.register_series_accessor("tab")
class TabSeriesAccessor(BaseTabAccessor):

    def __call__(
            self,
            x: t.Union[ColRef],
            y: t.Optional[ColRef] = None,
            values: t.Optional[pd.Series] = None,
            aggfunc: t.Optional[t.Union[ArrCallable, t.Iterable[ArrCallable]]] = None,
            **kwargs
    ) -> pd.DataFrame:
        x = self._lookup(x)
        values = self._lookup(values)
        return tabulate_one_way(x, values=values, aggfunc=aggfunc, **kwargs)



import os
import click


DEFAULT_JUPYTER_STARTUP_SCRIPT_NAME = "50-pandas_tab_init.py"
PANDAS_TAB_SCRIPTS_DIR = os.path.join(os.path.dirname(__file__), "scripts")
PANDAS_TAB_STARTUP_SCRIPT = "jupyter_init.py.jinja"


def get_ipython_startup_dir(profile_name: str) -> str:
    """Get the path to place startup scripts in."""
    from IPython.paths import get_ipython_dir
    from IPython.core.profiledir import ProfileDir
    profile_dir = ProfileDir.find_profile_dir_by_name(get_ipython_dir(), profile_name)
    return profile_dir.startup_dir


def create_jinja_env():
    """Lazy load jinja2 environment for scripts."""
    import jinja2
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(PANDAS_TAB_SCRIPTS_DIR)
    )


@click.group("pandas-tab")
def cli():
    """Manage IPython startup script for pandas_tab."""
    try:
        import IPython  # noqa: F401
        import jinja2  # noqa: F401
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            f"No module named {e.name!r}."
            " The Pandas-Tab CLI requires IPython and jinja2 to be installed."
            " You can `pip install pandas_tab[full]` to install these dependencies."
        )


@cli.command("init")
@click.option("--filename", "-f",
              type=click.STRING,
              default=DEFAULT_JUPYTER_STARTUP_SCRIPT_NAME)
@click.option("--profile-name", "-p",
              type=click.STRING,
              default="default",
              help="The name of the IPython profile to put the script in.")
@click.option("--noisy",
              is_flag=True,
              help="If set, the startup script will warn the user if pandas_tab"
                   " fails to load. By default, if pandas_tab fails to load"
                   "automatically in IPython, it will fail silently.")
@click.option("--overwrite",
              is_flag=True,
              help="Overwrite the script file if it already exists.")
def init(
        filename: str = DEFAULT_JUPYTER_STARTUP_SCRIPT_NAME,
        profile_name: str = "default",
        noisy: bool = False,
        overwrite: bool = False
):
    """Create startup script for IPython."""
    startup_dir = get_ipython_startup_dir(profile_name)
    script_file_path = os.path.join(startup_dir, filename)

    jinja_env = create_jinja_env()
    script = jinja_env.get_template("jupyter_init.py.jinja").render(noisy=noisy)

    if not overwrite and os.path.exists(script_file_path):
        raise FileExistsError(f"File {script_file_path} already exists.")

    with open(script_file_path, "w+") as f:
        f.write(script)

    click.echo(f"Created file {filename!r} in {startup_dir!r}")


@cli.command("delete")
@click.option("--filename", "-f",
              type=click.STRING,
              default=DEFAULT_JUPYTER_STARTUP_SCRIPT_NAME)
@click.option("--profile-name", "-p",
              type=click.STRING,
              default="default",
              help="The name of the IPython profile to look for the script in.")
def delete(
        filename: str = DEFAULT_JUPYTER_STARTUP_SCRIPT_NAME,
        profile_name: str = "default"
):
    """Delete startup script for IPython."""
    startup_dir = get_ipython_startup_dir(profile_name)
    script_file_path = os.path.join(startup_dir, filename)
    os.remove(script_file_path)
    click.echo(f"File {filename!r} deleted from {startup_dir}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re

from setuptools import setup
from setuptools import find_packages


def get_version(package):
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_long_description():
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="pandas_tab",
    python_requires=">=3.7",
    version=get_version("pandas_tab"),
    license="MIT",
    description="Data exploration done quick.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="ryxcommar",
    author_email="ryxcommar@gmail.com",
    url="https://github.com/ryxcommar/pandas_tab",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    entry_points={
        "console_scripts": {
            "pandas_tab = pandas_tab.__main__"
        }
    },
    install_requires=[
        "pandas>=0.23.0",
        "click"
    ],
    extras_require={
        "full": [
            "jinja2",
            "IPython"
        ]
    },
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Plugins",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
)
